#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "preview.h"
/*
	Cylinder, Sphere, and Cone drawing routines have been 
	adapted from VOGLE.
----
sumant
*/
extern int circle_precision;

fgetline(s, n, fp)
char  *s;
int  n;
FILE  *fp;
{
        register char  *cp = s;
        register int  c = EOF;

        while (--n > 0 && (c = getc(fp)) != EOF) {
                if (c == '\n' && (cp == s || cp[-1] != '\\'))
                        break;
                *cp++ = c;
        }
        *cp = '\0';
}

readobj_n_draw(input)
char  *input;
{
	FILE  *infp;
	char  buf[512];
	register int  c;

	if (input == NULL) {
		infp = stdin;
		input = "standard input";
	} else if (input[0] == '!') {
		if ((infp = popen(input+1, "r")) == NULL) {
			error("cannot execute",input);
		}
	} else if ((infp = fopen(input, "r")) == NULL) {
		return(0);
	}
	while ((c = getc(infp)) != EOF) {
		if (isspace(c))
			continue;
		if (c == '#') {				/* comment */
			fgets(buf, sizeof(buf), infp);
		} else if (c == '!') {			/* command */
			ungetc(c, infp);
			fgetline(buf, sizeof(buf), infp);
			readobj_n_draw(buf);
		} else {				/* object */
			ungetc(c, infp);
			get_n_drawobject(input, infp);
		}
	}
	if (input[0] == '!')
		pclose(infp);
	else
		fclose(infp);
	return(1);
}

/* Primary Type */
#define UNKNOWN -1
#define GEOMETRY 0
#define MATERIAL 1
/* Secondary Type */
#define POLYGON 0
#define SPHERE 1
#define BUBBLE 2
#define CYLINDER 3
#define CONE 4
#define TUBE 5
#define CUP 6
#define RING 7

#define LIGHT 0
#define ILLUM 1
#define GLOW 2
#define SPOTLIGHT 3
#define PLASTIC 4
#define METAL 5
#define TRANS 6
#define DIELECTRIC 7
#define INTERFACE 8
#define GLASS 9
#define PLASFUNC 10
#define METFUCH 11
#define METDATA 12


read_n_drawcone(name,fp)
char *name;
FILE *fp;
{
	read_n_drawcone_or_cylinder(name,fp,CONE);
}
read_n_drawcylinder(name,fp)
char *name;
FILE *fp;
{
	read_n_drawcone_or_cylinder(name,fp,CYLINDER);
}
int read_n_drawpolygon();
int read_n_drawsphere();
int read_n_drawring();
#define MAXPRIMITIVES 8
int (*read_n_draw[MAXPRIMITIVES])()={
	read_n_drawpolygon,
	read_n_drawsphere,
	read_n_drawsphere,
	read_n_drawcylinder,
	read_n_drawcone,
	read_n_drawcylinder,
	read_n_drawcone,
	read_n_drawring
};
int getindex(table,size,s)
char *table[];
int size;
char *s;
{
	int i;
	if (table==NULL) return(-1);
	for(i=0;i<size;i++){
		if (!strcmp(table[i],s)) return(i);
	}
	return(-1);
}

char *surface_table[MAXPRIMITIVES]={
	"polygon","sphere","bubble","cylinder","cone","tube","cup","ring"
};
int surface_table_size=MAXPRIMITIVES;

char *material_table[]={
	"light","illum","glow","spotlight","plastic",
	"metal","trans","dielectric","interface",
	"glass","plasfunc","metfunc","metdata"
};
int material_table_size=13; 

get_otype(s,t1,t2)
char *s;
int *t1,*t2;
{
	*t2 = getindex(surface_table,surface_table_size,s);
	if (*t2 >= 0){*t1 = GEOMETRY; return;}
	*t2 = getindex(material_table,material_table_size,s);
	if (*t2 >= 0){*t1 = MATERIAL; return;}
	*t1 = UNKNOWN;
}

#define MAXSIZE 248
static int local_material_table_size=0;
char *local_material_table[MAXSIZE];

int get_object_color(s)
char *s;
{
	int index;
	if (!strcmp(s,"void"))return(WHITE);
	index=getindex(local_material_table,local_material_table_size,s);
	
	return(8+index);
}

save_material(index,s1,s2,buf)
int index;
char *s1,*s2,*buf;
{
	if(local_material_table_size < MAXSIZE){
		char *s;
		char tmp[80];
		float r,g,b;

		if (sscanf(buf,"%s%f%f%f",tmp,&r,&g,&b)!=4)
			return;

		s=(char *)malloc(strlen(s1)+1);
		strcpy(s,s1);
		local_material_table[local_material_table_size]=s;

		switch(index){
		case LIGHT:
		case ILLUM:
		case GLOW :
		case SPOTLIGHT: r /= 255.; g /= 255.; b /=255.;break;
		case GLASS: 
		case INTERFACE:
		case TRANS:
			r = (1.-r);
			g = (1.-g);
			b = (1.-b);
			break;
		default : 
			break;
		}

		setcolor(8+local_material_table_size,r,g,b);
		local_material_table_size++;
	}
}

get_n_drawobject(name, fp)			/* read the next object */
char  *name;
FILE  *fp;
{
	char  buf[512];
	char  sbuf1[80],sbuf2[80],sbuf3[80];
	int   primary_type,secondary_type;

	fgets(buf, sizeof(buf), fp);
	sscanf(buf, "%s%s%s", sbuf1,sbuf2,sbuf3);
	/*
	   sbuf1 : Contains Modifier.
	   sbuf2 : Contains Type.
	   sbuf3 : Contains Identifies.
	*/
	get_otype(sbuf2,&primary_type,&secondary_type);
	fgets(buf, sizeof(buf), fp);
	fgets(buf, sizeof(buf), fp);
	switch(primary_type){
	case GEOMETRY :{
		int cindex= get_object_color(sbuf1);
		color(cindex);
		read_n_draw[secondary_type](name,fp);
	}
	break;
	case MATERIAL : {
		fgets(buf, sizeof(buf), fp);
		save_material(secondary_type,sbuf3,sbuf1,buf);
	}
	break;
	default : 
		fgets(buf, sizeof(buf), fp);
	break;
	}
}

read_n_drawsphere(name,fp)
char *name;
FILE *fp;
{
	float	i, x, y, z, r, a;
	float	delta = 180.0 / circle_precision;
	
	check(name,fp,4);
	if(fscanf(fp,"%g%g%g%g", &x, &y, &z, &r)!=4)
                error("Expected to read 4 numbers","read_n_drawsphere");

	pushmatrix();
		translate(x, y, z);
		/*
		 * create the longitudinal rings
		 */
		for (i = 0; i < 180; i += delta) {
			pushmatrix();
				rotate((float)i, 'y');
				circle(r);
			popmatrix();
		}
		/*
		 * create the latitudinal rings
		 */
		pushmatrix();
			rotate(90.0, 'x');
			for (i = -90.0; i < 90.0; i += delta) {
				a = r * cos((double)i * DTOR);
				z = r * sin((double)i * DTOR);
				pushmatrix();
					translate(0.0, 0.0, -z);
					circle(a);
				popmatrix();	
			}
		popmatrix();
	popmatrix();

}

read_n_drawcone_or_cylinder(name,fp,obj)
char *name;
FILE *fp;
char obj;
{
	float	apex_x, apex_y, apex_z, apex_r;
	float	base_x, base_y, base_z, base_r;
	float	len0, len1,  len_x, len_y, len_z;
	float	cosine, sine, dx1, dy1, dx2, dy2, delta;
	float	cx1, cy1, cx2, cy2;
	int	i;
	Matrix	m;

	if (obj==CONE){
	check(name,fp,8);
	if(fscanf(fp,"%g%g%g%g%g%g%g%g", 
			&base_x, &base_y, &base_z, 
			&apex_x, &apex_y, &apex_z, 
			&base_r, &apex_r)!=8)
                error("Expected to read 8 numbers","read_n_drawcone_or_cylinder");
	}
	else if (obj==CYLINDER){
		check(name,fp,7);
	if(fscanf(fp,"%g%g%g%g%g%g%g", 
			&base_x, &base_y, &base_z, 
			&apex_x, &apex_y, &apex_z, 
			&base_r)!=7)
                error("Expected to read 7 numbers","read_n_drawcone_or_cylinder");
		apex_r=base_r;
	}		

	len_x = apex_x - base_x;
	len_y = apex_y - base_y;
	len_z = apex_z - base_z;

	if ((len0 = sqrt(len_x * len_x + len_y * len_y + len_z * len_z)) < 1.0e-20)
		warn("c: zero length cone or cylinder",name);

	len1 = sqrt(len_x * len_x + len_z * len_z);

	move(base_x, base_y, base_z);
	draw(apex_x, apex_y, apex_z);

	pushmatrix();

		translate(base_x, base_y, base_z);

		identmatrix(m);

		if (len1 == 0.0) {
			cosine = 0.0;
			sine = 1.0;
		} else {
			cosine = len_z / len1;
			sine = len_x / len1;
		}

		/* rotate about y */
		m[0][0] = cosine;
		m[0][2] = -sine;
		m[2][0] = sine;
		m[2][2] = cosine;
		multmatrix(m);

		identmatrix(m);

		if (len0 == 0.0) {
			cosine = 0.0;
			sine = 1.0;
		} else {
			cosine = len1 / len0;
			sine = -len_y / len0;
		}

		/* rotate about x */
		m[1][1] = cosine;
		m[1][2] = sine;
		m[2][1] = -sine;
		m[2][2] = cosine;
		multmatrix(m);

		/*
		 * Draw the end circles...
		 */
		circle (base_r);
		pushmatrix();
			translate(0.0, 0.0, len0);
			circle (apex_r);
		popmatrix();



		/*
		 * Draw the logitudinal lines...
		 */
		delta = 2 * PI / circle_precision;

		cosine = cos(delta);
		sine = sin(delta);

		cx1 = base_r;
		cy1 = 0.0;
		cx2 = apex_r;
		cy2 = 0.0;

		move(cx1, cy1, 0.0);
		draw(cx2, cy2, len0);

		for (i = 0; i < circle_precision; i++) {
			dx1 = cx1;
			dy1 = cy1;
			cx1 = dx1 * cosine - dy1 * sine;
			cy1 = dx1 * sine + dy1 * cosine;
			dx2 = cx2;
			dy2 = cy2;
			cx2 = dx2 * cosine - dy2 * sine;
			cy2 = dx2 * sine + dy2 * cosine;
			move(cx1, cy1, 0.0);
			draw(cx2, cy2, len0);
		}
	popmatrix();
}

read_n_drawring(name,fp,obj)
char *name;
FILE *fp;
char obj;
{
	float 	center_x, center_y, center_z,
		normal_x, normal_y, normal_z;
	float	r,r0,r1,delta_r;
	float   len0,len1;
	float	cosine, sine;
	int	i;
	Matrix	m;

	check(name,fp,8);
	if(fscanf(fp,"%g%g%g%g%g%g%g%g", 
			&center_x, &center_y, &center_z, 
			&normal_x,&normal_y,&normal_z,
			&r0, &r1)!=8)
                error("Expected to read 8 numbers","read_n_drawring");

	len0 = sqrt(normal_x*normal_x + normal_y*normal_y + normal_z*normal_z);
	if (len0 < Small) error("Zero normal","read_n_drawring");
	len1 = sqrt(normal_x*normal_x+normal_z*normal_z);

	/* 
	 *	Assert r1 is the outer radius and r0 inner radius.
	 */
	if (r1 < r0) {
		r = r0; r0 = r1; r1 = r;
	}

	pushmatrix();

		translate(center_x,center_y,center_z);

		identmatrix(m);

		if (len1 == 0.0){
			cosine = 0.0;
			sine = 1.0;
		}
		else {
			cosine = normal_z/len1;
			sine = normal_x/len1;
		}

		/* rotate about y */
		m[0][0] = cosine;
		m[0][2] = -sine;
		m[2][0] = sine;
		m[2][2] = cosine;
		multmatrix(m);

		identmatrix(m);

		if (len0 == 0.0) {
			cosine = 0.0;
			sine = 1.0;
		} else {
			cosine = len1 / len0;
			sine = -normal_y / len0;
		}

		/* rotate about x */
		m[1][1] = cosine;
		m[1][2] = sine;
		m[2][1] = -sine;
		m[2][2] = cosine;
		multmatrix(m);

		/*
		 * Draw  circles...
		 */
		delta_r = (r1-r0)/circle_precision;
		for(i=0,r=r0;i<circle_precision;i++,r+=delta_r) circle (r);
	popmatrix();
}
/*
 * read_poly
 *
 * 	Read in and draw a polygon.
 */
read_n_drawpolygon(name,fp)
char *name;
FILE *fp;
{
	float	x, y, z, xp, yp, zp, x0, y0, z0;
	int	i, nv;

	if ((fscanf(fp, "%d\n", &nv) != 1)||((nv /=3) < 3))
		error("wanted an > 3 for verticies",name);
	
	read_three(name,fp,"%g %g %g\n", &x0, &y0, &z0);

	move(x0, y0, z0);

	xp = x0;
	yp = y0;
	zp = z0;

	for (i = 1; i < nv; i++) {
		read_three(name,fp,"%g %g %g\n", &x, &y, &z);

		draw(x, y, z);
		if (xp == x && yp == y && zp == z) {
			char buf[128];
			warn("p: point same as previous point",name);
			sprintf(buf, "%g %g %g = %g %g %g", xp, yp, zp, x, y, z);
			warn(buf,name);
		}

		xp = x;
		yp = y;
		zp = z;
	}
	/*
	 * Connect last to first...
	 */
	draw(x0, y0, z0);
}

error(s,name)
char	*s,*name;
{
	fprintf(stderr, "%s: %s\n", name, s);
	exit(1);
}
warn(s,name)
char	*s,*name;
{
	fprintf(stderr, "%s: %s\n", name, s);
}

/*
 * We seem to be reading lots of sets of 3 or 4 or 6 numbers....
 */
read_three(name,fp,f, a, b, c)
	char *name;
	FILE *fp;
	char	*f;
	float	*a, *b, *c;
{
	int	n;

	if ((n = fscanf(fp, f, a, b, c)) != 3) {
		char buf[128];
		sprintf(buf,"expected to read 3 numbers with '%s' format but got %d\n",
			f, n);
		error(buf,name);
	}

}
check(name,fp,num)
char *name;
FILE *fp;
int num;
{
	int n,m;
	if ((n = fscanf(fp,"%d",&m))!=1) error("Integer Expected",name);
	if (m!=num) error("Mismatch in number of values expected",name);
}
